"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. All Rights Reserved. See LICENSE for details.

urls.py - Mapping of URL's to our various views. Note we always used NAMED
          views for simplicity in linking later on.
"""

from django.conf import settings
from django.conf.urls import url
from django.contrib.auth.decorators import login_required
from django.views.generic import TemplateView
import django.contrib.auth.views

from helpdesk import settings as helpdesk_settings
from helpdesk.views import feeds, staff, public, kb
import helpdesk.views.api

class DirectTemplateView(TemplateView):
    extra_context = None
    def get_context_data(self, **kwargs):
        context = super(self.__class__, self).get_context_data(**kwargs)
        if self.extra_context is not None:
            for key, value in self.extra_context.items():
                if callable(value):
                    context[key] = value()
                else:
                    context[key] = value
        return context

urlpatterns = [
#'helpdesk.views.staff',
    url(r'^dashboard/$',
        staff.dashboard,
        name='helpdesk_dashboard'),

    url(r'^tickets/$',
        staff.ticket_list,
        name='helpdesk_list'),

    url(r'^tickets/update/$',
        staff.mass_update,
        name='helpdesk_mass_update'),

    url(r'^tickets/submit/$',
        staff.create_ticket,
        name='helpdesk_submit'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/$',
        staff.view_ticket,
        name='helpdesk_view'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/followup_edit/(?P<followup_id>[0-9]+)/$',
        staff.followup_edit,
        name='helpdesk_followup_edit'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/followup_delete/(?P<followup_id>[0-9]+)/$',
        staff.followup_delete,
        name='helpdesk_followup_delete'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/edit/$',
        staff.edit_ticket,
        name='helpdesk_edit'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/update/$',
        staff.update_ticket,
        name='helpdesk_update'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/delete/$',
        staff.delete_ticket,
        name='helpdesk_delete'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/hold/$',
        staff.hold_ticket,
        name='helpdesk_hold'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/unhold/$',
        staff.unhold_ticket,
        name='helpdesk_unhold'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/cc/$',
        staff.ticket_cc,
        name='helpdesk_ticket_cc'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/cc/add/$',
        staff.ticket_cc_add,
        name='helpdesk_ticket_cc_add'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/cc/delete/(?P<cc_id>[0-9]+)/$',
        staff.ticket_cc_del,
        name='helpdesk_ticket_cc_del'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/dependency/add/$',
        staff.ticket_dependency_add,
        name='helpdesk_ticket_dependency_add'),

    url(r'^tickets/(?P<ticket_id>[0-9]+)/dependency/delete/(?P<dependency_id>[0-9]+)/$',
        staff.ticket_dependency_del,
        name='helpdesk_ticket_dependency_del'),
        
    url(r'^tickets/(?P<ticket_id>[0-9]+)/attachment_delete/(?P<attachment_id>[0-9]+)/$',
        staff.attachment_del,
        name='helpdesk_attachment_del'),

    url(r'^raw/(?P<type>\w+)/$',
        staff.raw_details,
        name='helpdesk_raw'),

    url(r'^rss/$',
        staff.rss_list,
        name='helpdesk_rss_index'),

    url(r'^reports/$',
        staff.report_index,
        name='helpdesk_report_index'),

    url(r'^reports/(?P<report>\w+)/$',
        staff.run_report,
        name='helpdesk_run_report'),

    url(r'^save_query/$',
        staff.save_query,
        name='helpdesk_savequery'),

    url(r'^delete_query/(?P<id>[0-9]+)/$',
        staff.delete_saved_query,
        name='helpdesk_delete_query'),

    url(r'^settings/$',
        staff.user_settings,
        name='helpdesk_user_settings'),

    url(r'^ignore/$',
        staff.email_ignore,
        name='helpdesk_email_ignore'),

    url(r'^ignore/add/$',
        staff.email_ignore_add,
        name='helpdesk_email_ignore_add'),

    url(r'^ignore/delete/(?P<id>[0-9]+)/$',
        staff.email_ignore_del,
        name='helpdesk_email_ignore_del'),
]

urlpatterns += [
#'helpdesk.views.public',
    url(r'^$',
        public.homepage,
        name='helpdesk_home'),

    url(r'^view/$',
        public.view_ticket,
        name='helpdesk_public_view'),

    url(r'^change_language/$',
        public.change_language,
        name='helpdesk_public_change_language'),        
]

urlpatterns += [
    url(r'^rss/user/(?P<user_name>[\.A-Za-z0-9_\-\@]+)/$',
        login_required(feeds.OpenTicketsByUser()),
        name='helpdesk_rss_user'),
    
    url(r'^rss/user/(?P<user_name>[\.A-Za-z0-9_\-\@]+)/(?P<queue_slug>[A-Za-z0-9_-]+)/$',
        login_required(feeds.OpenTicketsByUser()),
        name='helpdesk_rss_user_queue'),
    
    url(r'^rss/queue/(?P<queue_slug>[A-Za-z0-9_\-\@]+)/$',
        login_required(feeds.OpenTicketsByQueue()),
        name='helpdesk_rss_queue'),
    
    url(r'^rss/unassigned/$',
        login_required(feeds.UnassignedTickets()),
        name='helpdesk_rss_unassigned'),
    
    url(r'^rss/recent_activity/$',
        login_required(feeds.RecentFollowUps()),
        name='helpdesk_rss_activity'),   
]


urlpatterns += [
    url(r'^api/(?P<method>[a-z_-]+)/$',
        helpdesk.views.api.api,
        name='helpdesk_api'),

    url(r'^login/$',
        django.contrib.auth.views.login,
        {'template_name': 'helpdesk/registration/login.html'},
        name='login'),

    url(r'^logout/$',
        django.contrib.auth.views.logout,
        {
            'template_name': 'helpdesk/registration/login.html',
            'next_page': '../'
        },
        name='logout'),
]

if helpdesk_settings.HELPDESK_KB_ENABLED:
    urlpatterns += [
    #'helpdesk.views.kb',
        url(r'^kb/$',
            kb.index, name='helpdesk_kb_index'),
        
        url(r'^kb/(?P<item>[0-9]+)/$',
            kb.item, name='helpdesk_kb_item'),

        url(r'^kb/(?P<item>[0-9]+)/vote/$',
            kb.vote, name='helpdesk_kb_vote'),

        url(r'^kb/(?P<slug>[A-Za-z0-9_-]+)/$',
            kb.category, name='helpdesk_kb_category'),
    ]

urlpatterns += [
    url(r'^api/$',
        TemplateView.as_view(template_name='helpdesk/help_api.html'),
        name='helpdesk_api_help'),

    url(r'^help/context/$',
        TemplateView.as_view(template_name='helpdesk/help_context.html'),
        name='helpdesk_help_context'),

        url(r'^system_settings/$',
            DirectTemplateView.as_view(
                template_name='helpdesk/system_settings.html',
                extra_context={
                    'ADMIN_URL': getattr(settings, 'ADMIN_URL', '/admin/')
                }),
            name='helpdesk_system_settings'),
    ]



"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. All Rights Reserved. See LICENSE for details.

lib.py - Common functions (eg multipart e-mail)
"""
import os
try:
    from base64 import urlsafe_b64encode as b64encode # pylint: disable=unused-import
except ImportError:
    from base64 import encodestring as b64encode # pylint: disable=unused-import
try:
    from base64 import urlsafe_b64decode as b64decode # pylint: disable=unused-import
except ImportError:
    from base64 import decodestring as b64decode # pylint: disable=unused-import

import logging

from django.template import engines
from django.utils.safestring import mark_safe
from django.utils.encoding import smart_str
from django.contrib.sites.models import Site
from django.conf import settings

logger = logging.getLogger('helpdesk')

chart_colours = ('80C65A', '990066', 'FF9900', '3399CC', 'BBCCED', '3399CC', 'FFCC33')

def send_templated_mail(template_name, email_context, recipients, sender=None, bcc=None, fail_silently=False, files=None):
    """
    send_templated_mail() is a warpper around Django's e-mail routines that
    allows us to easily send multipart (text/plain & text/html) e-mails using
    templates that are stored in the database. This lets the admin provide
    both a text and a HTML template for each message.

    template_name is the slug of the template to use for this message (see
        models.EmailTemplate)

    email_context is a dictionary to be used when rendering the template

    recipients can be either a string, eg 'a@b.com', or a list of strings.

    sender should contain a string, eg 'My Site <me@z.com>'. If you leave it
        blank, it'll use settings.DEFAULT_FROM_EMAIL as a fallback.

    bcc is an optional list of addresses that will receive this message as a
        blind carbon copy.

    fail_silently is passed to Django's mail routine. Set to 'True' to ignore
        any errors at send time.

    files can be a list of file paths to be attached, or it can be left blank.
        eg ('/tmp/file1.txt', '/tmp/image.png')

    """
    from django.conf import settings
    from django.core.mail import EmailMultiAlternatives
    from django.template import loader, Context

    from helpdesk.models import EmailTemplate
    
    context = Context(email_context)

    if hasattr(context['queue'], 'locale'):
        locale = getattr(context['queue'], 'locale', '')
    else:
        locale = context['queue'].get('locale', 'en')
    if not locale:
        locale = 'en'

    t = None
    try:
        t = EmailTemplate.objects.get(template_name__iexact=template_name, locale=locale)
    except EmailTemplate.DoesNotExist:
        try:
            t = EmailTemplate.objects.get(template_name__iexact=template_name, locale__isnull=True)
        except EmailTemplate.DoesNotExist:
            print('template "%s" does not exist, no mail sent' % template_name)
            return # just ignore if template doesn't exist

    if not sender:
        sender = settings.DEFAULT_FROM_EMAIL

    footer_file = os.path.join('helpdesk', locale, 'email_text_footer.txt')

    #text_part = loader.get_template_from_string("%s{%% include '%s' %%}" % (t.plain_text, footer_file)).render(context)
    text_part = engines['django'].from_string("%s{%% include '%s' %%}" % (t.plain_text, footer_file)).render(email_context)

    email_html_base_file = os.path.join('helpdesk', locale, 'email_html_base.html')

    if context.has_key('comment') and context['comment']:
        html_txt = context['comment']
        html_txt = html_txt.replace('\r\n', '<br>')
        context['comment'] = mark_safe(html_txt)

    html_part = engines['django'].from_string(
        "{%% extends '%s' %%}{%% block title %%}%s{%% endblock %%}{%% block content %%}%s{%% endblock %%}" \
            % (email_html_base_file, t.heading, t.html)
        ).render(email_context)

    subject_part = engines['django'].from_string("{{ ticket.ticket }} {{ ticket.title|safe }} %s" % t.subject).render(email_context)

    if isinstance(recipients, (str, unicode)):
        if recipients.find(','):
            recipients = recipients.split(',')
    elif not isinstance(recipients, list):
        recipients = [recipients]

    msg = EmailMultiAlternatives(
        subject_part,
        text_part,
        sender,
        recipients,
        bcc=bcc)
    msg.attach_alternative(html_part, "text/html")

    if files:
        if not isinstance(files, list):
            files = [files,]
        for f in files:
            msg.attach_file(f)

    return msg.send(fail_silently)


def query_to_dict(results, descriptions):
    """
    Replacement method for cursor.dictfetchall() as that method no longer
    exists in psycopg2, and I'm guessing in other backends too.

    Converts the results of a raw SQL query into a list of dictionaries, suitable
    for use in templates etc.
    """

    output = []
    for data in results:
        row = {}
        i = 0
        for column in descriptions:
            row[column[0]] = data[i]
            i += 1

        output.append(row)
    return output


def apply_query(queryset, params):
    """
    Apply a dict-based set of filters & parameters to a queryset.

    queryset is a Django queryset, eg MyModel.objects.all() or
             MyModel.objects.filter(user=request.user)

    params is a dictionary that contains the following:
        filtering: A dict of Django ORM filters, eg:
            {'user__id__in': [1, 3, 103], 'title__contains': 'foo'}
        other_filter: Another filter of some type, most likely a
            set of Q() objects.
        sorting: The name of the column to sort by
    """
    for key in params['filtering'].keys():
        kwargs = {key: params['filtering'][key]}
        queryset = queryset.filter(**kwargs)

    if params.get('other_filter', None):
        # eg a Q() set
        queryset = queryset.filter(params['other_filter'])

    sorting = params.get('sorting', None)
    if sorting:
        sortreverse = params.get('sortreverse', None)
        if sortreverse:
            sorting = "-%s" % sorting
        queryset = queryset.order_by(sorting)

    return queryset


def safe_template_context(ticket):
    """
    Return a dictionary that can be used as a template context to render
    comments and other details with ticket or queue parameters. Note that
    we don't just provide the Ticket & Queue objects to the template as
    they could reveal confidential information. Just imagine these two options:
        * {{ ticket.queue.email_box_password }}
        * {{ ticket.assigned_to.password }}

    Ouch!

    The downside to this is that if we make changes to the model, we will also
    have to update this code. Perhaps we can find a better way in the future.
    """

    context = {
        'queue': {},
        'ticket': {},
        }
    queue = ticket.queue

    for field in ('title', 'slug', 'email_address', 'from_address', 'locale'):
        attr = getattr(queue, field, None)
        if callable(attr):
            context['queue'][field] = attr()
        else:
            context['queue'][field] = attr

    for field in (
        'title', 'created', 'modified', 'submitter_email',
        'status', 'get_status_display', 'on_hold', 'description',
        'resolution', 'priority', 'get_priority_display',
        'last_escalation', 'ticket', 'ticket_for_url',
        'get_status', 'ticket_url', 'staff_url', '_get_assigned_to'
    ):
        attr = getattr(ticket, field, None)
        if callable(attr):
            context['ticket'][field] = '%s' % attr()
        else:
            context['ticket'][field] = attr

    context['ticket']['queue'] = context['queue']
    context['ticket']['assigned_to'] = context['ticket']['_get_assigned_to']

    return context


def text_is_spam(text, request):
    # Based on a blog post by 'sciyoshi':
    # http://sciyoshi.com/blog/2008/aug/27/using-akismet-djangos-new-comments-framework/
    # This will return 'True' is the given text is deemed to be spam, or
    # False if it is not spam. If it cannot be checked for some reason, we
    # assume it isn't spam.
    try:
        from helpdesk.akismet import Akismet
    except ImportError:
        return False
    try:
        site = Site.objects.get_current()
    except Site.DoesNotExist:
        site = Site(domain='configure-django-sites.com')

    ak = Akismet(
        blog_url='http://%s/' % site.domain,
        agent='django-helpdesk',
    )

    if hasattr(settings, 'TYPEPAD_ANTISPAM_API_KEY'):
        ak.setAPIKey(key=settings.TYPEPAD_ANTISPAM_API_KEY)
        ak.baseurl = 'api.antispam.typepad.com/1.1/'
    elif hasattr(settings, 'AKISMET_API_KEY'):
        ak.setAPIKey(key=settings.AKISMET_API_KEY)
    else:
        return False

    if ak.verify_key():
        ak_data = {
            'user_ip': request.META.get('REMOTE_ADDR', '127.0.0.1'),
            'user_agent': request.META.get('HTTP_USER_AGENT', ''),
            'referrer': request.META.get('HTTP_REFERER', ''),
            'comment_type': 'comment',
            'comment_author': '',
        }

        return ak.comment_check(smart_str(text), data=ak_data)

    return False



from __future__ import print_function

# Make warnings throw explicit exceptions so we're forced to fix them.
import warnings
import logging

from django.test import TestCase
from django.core import mail
from django.test.client import Client
from django.core.urlresolvers import reverse

#import six

from helpdesk.models import Queue, CustomField, Ticket, User, EmailTemplate

warnings.simplefilter('error', RuntimeWarning)

logger = logging.getLogger('helpdesk')
logger.setLevel(logging.DEBUG)

class TicketBasicsTestCase(TestCase):
    
    fixtures = [
        'initial_data.json',
        'test_data.json',
    ]
    
    def setUp(self):
        
        self.queue_public = Queue.objects.create(
            title='Queue 1',
            slug='q1',
            allow_public_submission=True,
            new_ticket_cc='new.public@example.com',
            updated_ticket_cc='update.public@example.com')
            
        self.queue_private = Queue.objects.create(
            title='Queue 2',
            slug='q2',
            allow_public_submission=False,
            new_ticket_cc='new.private@example.com',
            updated_ticket_cc='update.private@example.com')

        self.ticket_data = {
            'title': 'Test Ticket',
            'description': 'Some Test Ticket',
        }

        self.client = Client()

    def test_create_ticket_direct(self):
        email_count = len(mail.outbox)
        ticket_data = dict(queue=self.queue_public, **self.ticket_data)
        ticket = Ticket.objects.create(**ticket_data)
        self.assertEqual(ticket.ticket_for_url, "q1-%s" % ticket.id)
        self.assertEqual(email_count, len(mail.outbox))
        

    def test_create_ticket_public(self):
        email_count = len(mail.outbox)

        response = self.client.get(reverse('helpdesk_home'))
        self.assertEqual(response.status_code, 200)

        # Submit a ticket.
        post_data = {
            'title': 'Test ticket title',
            'queue': self.queue_public.id,
            'submitter_email': 'ticket1.submitter@example.com',
            'body': 'Test ticket body',
            'priority': 3,
        }
        response = self.client.post(
            reverse('helpdesk_home'), post_data, follow=True)
        last_redirect = response.redirect_chain[-1]
        last_redirect_url = last_redirect[0]
        last_redirect_status = last_redirect[1]
        
        # Ensure we landed on the "View" page.
        self.assertEqual(
            last_redirect_url.split('?')[0],
            reverse('helpdesk_public_view'))
            
        # Ensure submitter, new-queue + update-queue were all emailed.
#        for email in mail.outbox:
#            #print(dir(email), email.__dict__.keys())
#            print(email.subject, email.to)
        self.assertEqual(email_count+3, len(mail.outbox))
        
        email_templates = EmailTemplate.objects.filter(locale='en')
        self.assertEqual(email_templates.count(), 16)
        
        # Lookup ticket.
        t = Ticket.objects.all().order_by('-id')[0]
        
        users = User.objects.all()
        self.assertEqual(users.count(), 2)
        
        # Assign ticket.
        mail.outbox = []
        t.assigned_to = users[0]
        t.save(send_email=True)
        
        # Confirm a reassignment notice email was sent.
        self.assertEqual(len(mail.outbox), 1)
        #print(mail.outbox[0].body)
    
    def test_create_ticket_private(self):
        email_count = len(mail.outbox)
        post_data = {
            'title': 'Private ticket test',
            'queue': self.queue_private.id,
            'submitter_email': 'ticket2.submitter@example.com',
            'body': 'Test ticket body',
            'priority': 3,
        }

        response = self.client.post(reverse('helpdesk_home'), post_data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(email_count, len(mail.outbox))
        self.assertContains(response, 'Select a valid choice.')

    def test_create_ticket_customfields(self):
        email_count = len(mail.outbox)
        
        queue_custom = Queue.objects.create(
            title='Queue 3',
            slug='q3',
            allow_public_submission=True,
            updated_ticket_cc='update.custom@example.com')
            
        custom_field_1 = CustomField.objects.create(
            name='textfield',
            label='Text Field',
            data_type='varchar',
            max_length=100,
            ordering=10,
            required=False,
            empty_selection_list=False,
            staff_only=False)
            
        post_data = {
            'queue': queue_custom.id,
            'title': 'Ticket with custom text field',
            'submitter_email': 'ticket3.submitter@example.com',
            'body': 'Test ticket body',
            'priority': 3,
            'custom_textfield': 'This is my custom text.',
        }

        response = self.client.post(
            reverse('helpdesk_home'), post_data, follow=True)

        custom_field_1.delete()
        last_redirect = response.redirect_chain[-1]
        last_redirect_url = last_redirect[0]
        last_redirect_status = last_redirect[1]
        # Ensure we landed on the "View" page.
        self.assertEqual(
            last_redirect_url.split('?')[0],
            reverse('helpdesk_public_view'))
        # Ensure only two e-mails were sent - submitter & updated.
        self.assertEqual(email_count+2, len(mail.outbox))
        
    def test_helpdesk_submit(self):
        """
        Confirm a user can submit a ticket on the ticket submission page.
        """
        
        queue_simple = Queue.objects.create(
            title='Queue Simple',
            slug='queue-simple',
            allow_public_submission=True)
            
        users = User.objects.all()
        self.assertEqual(users.count(), 2)
        submitter = User.objects.get(username='jondoe')
        submitter.set_password('password')
        submitter.save()
        
        assigned_user = User.objects.get(username='janedoe')

        # We're not logged in yet, so we should be redirected.
        response = self.client.get(reverse('helpdesk_submit'))
        self.assertEqual(response.status_code, 302)
        
        # Login.
        #TODO:why doesn't this work?
#        response = self.client.post(
#            '/admin/login/',
#            {'username': 'jondoe', 'password': 'password'})
#        self.assertEqual(response.status_code, 200)
#        print(response)
#        self.assertTrue('Please log in again' not in str(response))
        self.client.login(username=submitter.username, password='password')
        
        # Now we should be able to load the submission page.
        response = self.client.get(reverse('helpdesk_submit'))
        self.assertEqual(response.status_code, 200)
        
#        # Submit a ticket.
        post_data = dict(
            queue=queue_simple.id,
            title='Test ticket title',
            submitter_email=submitter.email,
            body='Test ticket body',
            priority=3,
            assigned_to=assigned_user.id,
        )
        response = self.client.post(
            reverse('helpdesk_submit'), post_data, follow=True)
            
        # Ensure we landed on the ticket's page.
        last_redirect = response.redirect_chain[-1]
        last_redirect_url = last_redirect[0]
        last_redirect_status = last_redirect[1]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            last_redirect_url.split('?')[0],
            reverse('helpdesk_view', args=(1,)))
#            
#        # Ensure submitter, new-queue + update-queue were all emailed.
        for email in mail.outbox:
#            #print(dir(email), email.__dict__.keys())
            print(email.subject, email.to)
        self.assertEqual(len(mail.outbox), 2)
#        
#        email_templates = EmailTemplate.objects.filter(locale='en')
#        self.assertEqual(email_templates.count(), 16)
#        
#        # Lookup ticket.
#        t = Ticket.objects.all().order_by('-id')[0]
        

class PublicActionsTestCase(TestCase):
    """
    Tests for public actions:
    - View a ticket
    - Add a followup
    - Close resolved case
    """
    def setUp(self):
        """
        Create a queue & ticket we can use for later tests.
        """
        
        self.queue = Queue.objects.create(
            title='Queue 1',
            slug='q',
            allow_public_submission=True,
            new_ticket_cc='new.public@example.com',
            updated_ticket_cc='update.public@example.com')
            
        self.ticket = Ticket.objects.create(
            title='Test Ticket',
            queue=self.queue,
            submitter_email='test.submitter@example.com',
            description='This is a test ticket.')

        self.client = Client()

    def test_public_view_ticket(self):
        response = self.client.get(
            '%s?ticket=%s&email=%s' % (reverse('helpdesk_public_view'),
            self.ticket.ticket_for_url, 'test.submitter@example.com'))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateNotUsed(response, 'helpdesk/public_view_form.html')

    def test_public_close(self):
        old_status = self.ticket.status
        old_resolution = self.ticket.resolution
        resolution_text = 'Resolved by test script'

        ticket = Ticket.objects.get(id=self.ticket.id)
        
        ticket.status = Ticket.RESOLVED_STATUS
        ticket.resolution = resolution_text
        ticket.save()

        current_followups = ticket.followup_set.all().count()
        
        response = self.client.get(
            '%s?ticket=%s&email=%s&close' % (reverse('helpdesk_public_view'),
            ticket.ticket_for_url, 'test.submitter@example.com'))
        
        ticket = Ticket.objects.get(id=self.ticket.id)

        self.assertEqual(response.status_code, 302)
        self.assertTemplateNotUsed(response, 'helpdesk/public_view_form.html')
        self.assertEqual(ticket.status, Ticket.CLOSED_STATUS)
        self.assertEqual(ticket.resolution, resolution_text)
        self.assertEqual(
            current_followups+1, ticket.followup_set.all().count())
        
        ticket.resolution = old_resolution
        ticket.status = old_status
        ticket.save()
